/*
  G. Wiesenekker. GWD a draughts program.
  Copyright (C) 1993-1994 G. Wiesenekker
  E-mail: wiesenekker@sara.nl

  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.

  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.

  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include "draughts.h"
#if MACHINE == DOS
#include "xms.h"
#define KB 1024
struct XMSmovestruct c2x;
struct XMSmovestruct x2c;
char XMSresult[4];
unsigned int XMShandle;
unsigned long XMSoffset;
#endif
static char *FILE_NAME = __FILE__;
int dummy = 0;

int DepthMax = 32;
int TimeLimit = 30;
int TimeExtra = 30;
int Append = TRUE;
int GameTimeUsed = 0;
int UsePos = TRUE;
int UseEndgame = TRUE;
int my_colour, your_colour;

enum { X1 =       6,  X2,  X3,  X4,  X5};
enum { X6 =  X1 + 5,  X7,  X8,  X9, X10};
enum {X11 =  X6 + 6, X12, X13, X14, X15};
enum {X16 = X11 + 5, X17, X18, X19, X20};
enum {X21 = X16 + 6, X22, X23, X24, X25};
enum {X26 = X21 + 5, X27, X28, X29, X30};
enum {X31 = X26 + 6, X32, X33, X34, X35};
enum {X36 = X31 + 5, X37, X38, X39, X40};
enum {X41 = X36 + 6, X42, X43, X44, X45};
enum {X46 = X41 + 5, X47, X48, X49, X50};

byte_t map[50] =
{
	 X1,  X2,  X3,  X4,  X5,
	 X6,  X7,  X8,  X9, X10,
	X11, X12, X13, X14, X15,
	X16, X17, X18, X19, X20,
	X21, X22, X23, X24, X25,
	X26, X27, X28, X29, X30,
	X31, X32, X33, X34, X35,
	X36, X37, X38, X39, X40,
	X41, X42, X43, X44, X45,
	X46, X47, X48, X49, X50
};

byte_t mirror_map[50] =
{
	X50, X49, X48, X47, X46,
	X45, X44, X43, X42, X41,
	X40, X39, X38, X37, X36,
	X35, X34, X33, X32, X31,
	X30, X29, X28, X27, X26,
	X25, X24, X23, X22, X21,
	X20, X19, X18, X17, X16,
	X15, X14, X13, X12, X11,
	X10,  X9,  X8,  X7,  X6,
	 X5,  X4,  X3,  X2,  X1
};

int inverse_map[BOARD_MAX] =
{
	-1, -1, -1, -1, -1, -1,
		 0,  1,  2,  3,  4,
	 5,  6,  7,  8,  9, -1,
	  10, 11, 12, 13, 14,
	15, 16, 17, 18, 19, -1,
	  20, 21, 22, 23, 24,
	25, 26, 27, 28, 29, -1,
	  30, 31, 32, 33, 34,
	35, 36, 37, 38, 39, -1,
	  40, 41, 42, 43, 44,
	45, 46, 47, 48, 49, -1,
	  -1, -1, -1, -1, -1
};

int pat_map[BOARD_MAX] =
{
	 0,  0,  0,  0,  0,  0,
		 1,  2,  3,  4,  5,
	 6,  7,  8,  9, 10,  0,
	  11, 12, 13, 14, 15,
	16, 17, 18, 19, 20,  0,
	  21, 22, 23, 24, 25,
	26, 27, 28, 29, 30,  0,
	  31, 32, 33, 34, 35,
	36, 37, 38, 39, 40,  0,
	  41, 42, 43, 44, 45,
	46, 47, 48, 49, 50,  0,
	   0,  0,  0,  0,  0
};

char nota[3];
#define GET_NOTA(X) (sprintf(nota, "%d", (int) pat_map[X]), nota)

byte_t row_table[50] =
{
	  9, 9, 9, 9, 9,
	8, 8, 8, 8, 8,
	  7, 7, 7, 7, 7,
	6, 6, 6, 6, 6,
	  5, 5, 5, 5, 5,
	4, 4, 4, 4, 4,
	  3, 3, 3, 3, 3,
	2, 2, 2, 2, 2,
	  1, 1, 1, 1, 1,
	0, 0, 0, 0, 0
};
int *row;

char nota_colour[4] = {'w', 'b', '.', '?'};
char *colour_as_string[2] = {"white ", "black "};

char nota_kind[2] = {'O', 'X'};

#define UPDATE_KEY(X, Y) {(X).lo ^= (Y).lo; (X).hi ^= (Y).hi;}

hash_key_t *hash_key[2][2];
hash_key_t *threat_key;

#define SCORE_INVAL   32000
#define SCORE_LIMIT   28000
#define SCORE_MATE    24000
#define SCORE_ABORT   23000
#define SCORE_ENDGAME 20000

int value[2] = 
{
	MAN_SCORE, CROWN_SCORE
};

byte_t man_visit_table[50] = 
{
	   0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,
	   0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,
	   0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,
	   0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0,
	   0,  0,  0,  0,  0,
	 0,  0,  0,  0,  0
};

int *white_man_visit, *black_man_visit;
int *man_visit[2];

#ifdef SMALL
#define MOVE_MAX  512
#define STACK_MAX 2048
#define VAR_MAX   128
#else
#define MOVE_MAX  1024
#define STACK_MAX 4096
#define VAR_MAX   256
#endif
#define CAPT_MAX  (2 * LIST_MAX)

move_t **move_stack;
move_t **move_top;
move_t *stack;
move_t *stack_top;
board_t **capt_stack;
board_t **capt_top;
var_t *var_stack;

int white_man_dir[5] =
	{-6, -5, 5, 6, 0};
int black_man_dir[5] =
	{5, 6, -6, -5, 0};
int *man_dir[2] =
	{white_man_dir, black_man_dir};

int white_crown_dir[5] =
	{-6, -5, 5, 6, 0};
int black_crown_dir[5] =
	{5, 6, -6, -5, 0};
int *crown_dir[2] =
	{white_crown_dir, black_crown_dir};

int *white_dir[2] =
	{white_man_dir, white_crown_dir};
int *black_dir[2] =
	{black_man_dir, black_crown_dir};
int **dir[2] =
	{white_dir, black_dir};
	
int prom_row[2] = {9, 0};

pat_t white_pat[2] = {BIT(WBIT) | BIT(MBIT) | BIT(WMBIT), BIT(WBIT) | BIT(CBIT) | BIT(WCBIT)};
pat_t black_pat[2] = {BIT(BBIT) | BIT(MBIT) | BIT(BMBIT), BIT(BBIT) | BIT(CBIT) | BIT(BCBIT)};
pat_t *pat[2] = {white_pat, black_pat};

board_t *board[BOARD_MAX];
pat_t asm_board[BOARD_MAX];
pat_t b[51];
pat_t bempty[51];
board_t empty = {NONE, INVALID, 0, INVALID, NULL, NULL, NULL};
board_t side = {SIDE, INVALID, 0, INVALID, NULL, NULL, NULL};

board_t white_man_list[LIST_MAX], black_man_list[LIST_MAX];
board_t *man_list[2] =
	{white_man_list, black_man_list};

board_t white_crown_list[LIST_MAX], black_crown_list[LIST_MAX];
board_t *crown_list[2] =
	{white_crown_list, black_crown_list};

board_t *white_list[2] =
	{white_man_list, white_crown_list};
board_t *black_list[2] =
	{black_man_list, black_crown_list};
board_t **list[2] =
	{white_list, black_list};
	
int white_piece_count[2];
int black_piece_count[2];
int	*piece_count[2] = {white_piece_count, black_piece_count};
int tot_count[2];

#define NODE_MAX 256

#ifdef SMALL
#define CACHE_BITS       8
#define SCORE_CACHE_BITS 8
#else
#if MACHINE == MAC
#define CACHE_BITS       13
#define SCORE_CACHE_BITS 13
#elif MACHINE == DOS
#define CACHE_BITS       12
#define SCORE_CACHE_BITS 12
#else
#define CACHE_BITS       14
#define SCORE_CACHE_BITS 14
#endif
#endif

#define CACHE_MASK       (~(~0L << CACHE_BITS))
#define SCORE_CACHE_MASK (~(~0L << SCORE_CACHE_BITS))

#ifdef SMALL
#define CACHE_DEPTH_BITS 4
#else
#define CACHE_DEPTH_BITS 6
#endif
#define CACHE_DEPTH_MASK (~(~0UL << CACHE_DEPTH_BITS))

#define CACHE_INVAL BITL(CACHE_DEPTH_BITS)
#define CACHE_LOWER BITL(CACHE_DEPTH_BITS + 1)
#define CACHE_UPPER BITL(CACHE_DEPTH_BITS + 2)
#define CACHE_TRUE  BITL(CACHE_DEPTH_BITS + 3)

#define CACHE_KIND_BITS (CACHE_DEPTH_BITS + 4)
#define CACHE_KIND_MASK (~(~0L << CACHE_KIND_BITS))

#if CACHE_BITS < CACHE_KIND_BITS
#error CACHE_BITS should not be smaller than CACHE_KIND_BITS
#endif

node_t *node_stack, *node;
int inode;

typedef struct
{
	int cache_best;
	score_t cache_score;
	hash_key_t cache_key;
} cache_t;

typedef struct
{
	score_t cache_score;
	hash_key_t cache_key;
} score_cache_t;

#if MACHINE == DOS
cache_t huge *cache[2];
score_cache_t huge *score_cache[2];
#else
cache_t *cache[2];
score_cache_t *score_cache[2];
#endif

int null_move, principle_variation;
int nthreats;

long ngens, ndos;
long nevals, nhits, nextends;
long nodes_tot, nevals_tot;

clock_t clock_begin, clock_end;
double time_used;

char *game_moves;

void display_move(char *, move_t *, char *);
void display_move(char *prefix, move_t *m, char *postfix)
{
	my_printf("%s%s", prefix, GET_NOTA(*m));
	m++;
	while(*m != 0)
	{
		my_printf("-%s", GET_NOTA(*m));
		m++;
	}
	my_printf("%s", postfix);
}

char move_string[STRING_MAX];

void move_to_string(move_t *m, int option)
{
	move_t i, j;
	
	i = *m++;
	j = *m++;
	if (*m == 0)
	{
		strcpy(move_string, GET_NOTA(i));
		strcat(move_string, "-");
		strcat(move_string, GET_NOTA(j));
	}
	else
	{
		strcpy(move_string, GET_NOTA(i));
		strcat(move_string, "x");
		strcat(move_string, GET_NOTA(j));
		if (option)
		{
			while(*m != 0)
			{
				strcat(move_string, "x");
				strcat(move_string, GET_NOTA(*m));
				m++;
			}
		}
	}
}

void display_board(void)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 10; j++)
	{
		if ((j & 1) == 0) my_printf("  ");
		for (i = 0; i < 5; i++)
		{
			if (board[map[k]] == &empty)
				my_printf("  ..");
			else
				my_printf("  %c%c",
					nota_colour[board[map[k]]->board_colour],
					nota_kind[board[map[k]]->board_kind]);
			k++;
		}
		my_printf("\n");
	}
}

void write_board(FILE *f)
{
	int i, j, k;
	
	k = 0;
	for (j = 0; j < 10; j++)
	{
		if ((j & 1) == 0) my_printf("  ");
		for (i = 0; i < 5; i++)
		{
			if (board[map[k]] == &empty)
				fprintf(f, "  ..");
			else
				fprintf(f, "  %c%c",
					nota_colour[board[map[k]]->board_colour],
					nota_kind[board[map[k]]->board_kind]);
			k++;
		}
		fprintf(f, "\n");
	}
}

void display_list(void)
{
	int icolour;
	int ikind;
	board_t *p;

	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		my_printf("%s%d", colour_as_string[icolour], tot_count[icolour]);
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			my_printf(" %d'%c'", piece_count[icolour][ikind], nota_kind[ikind]);
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where == INVALID)
					my_printf(" ??");
				else
					my_printf(" %s", GET_NOTA(p->board_where));
			}
		}
		my_printf("\n");
	}
}

int return_mat_score(int);
int return_mat_score(int my_colour)
{
	int sum;
	int ikind;
	board_t *p;
	
	sum = 0;
	for (ikind = MAN; ikind <= CROWN; ikind++)
	{
		for (p = list[my_colour][ikind]; p->board_kind != INVALID; p++)
			if (p->board_where != INVALID) sum += p->board_value;
	}
	return(sum);
}

hash_key_t return_cur_key(void);
hash_key_t return_cur_key(void)
{
	int icolour, ikind;
	hash_key_t key;
	board_t *p;
	
	key.lo = key.hi = 0UL;
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
					UPDATE_KEY(key, p->board_key[p->board_where])
			}
		}
	}
	return(key);
}

void update_key(hash_key_t *key)
{
	UPDATE_KEY(*key, threat_key[nthreats])
}

int check_list(void);
int check_list(void)
{
	int i;
	int icolour;
	board_t *p;
	int n, m;
	int ikind;
	
	for (i = 0; i < 50; i++)
		if ((board[map[i]] != &empty) and (board[map[i]]->board_where != map[i])) return(1);
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		n = 0;
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			m = 0;
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
				{
					if (board[p->board_where] != p) return(2);
					m++;
				}
			}
			if (m != piece_count[icolour][ikind]) return(3);
			n += m;
		}
		if (n != tot_count[icolour]) return(4);
	}
	return(0);
}

int check_asm_board(void);
int check_asm_board(void)
{
	int i;
	board_t *p;
	
	for (i = 0; i < 50; i++)
	{
		p = board[map[i]];
		if ((p == &empty) and (asm_board[map[i]] != 0)) return(1);
		if ((p != &empty) and (asm_board[map[i]] != p->board_pat)) return(2);
	}
	return(0);
}

void compact_lists(void)
{
	int icolour, ikind;
	board_t *p;
	int err;
	int n, i;
	board_t temp_list[LIST_MAX];
	
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			n = 0;
			for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
			{
				if (p->board_where != INVALID)
				{
					BUG(n >= LIST_MAX)
					temp_list[n] = *p;
					n++;
				}
			}
			p = list[icolour][ikind];
			for (i = 0; i < n; i++)
			{
				*p = temp_list[i];
				board[p->board_where] = p;
				p++;
			}
			for ( ; i < LIST_MAX; i++)
			{
				p->board_kind = p->board_where = INVALID;
				p++;
			}
		}
	}
	if ((err = check_list()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("list error")
	}
	if ((err = check_asm_board()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("asm error")
	}
}

int read_board(FILE *f)
{
	int i;
	int temp_colour[50], temp_kind[50];
	char temp[LINE_MAX];
	int icolour, ikind;
	board_t *p;
	int err;

	for (i = 0; i < 50; i++)
	{
		if (fscanf(f, "%2s", temp) != 1) return(FALSE);
		if (temp[0] == 'q') return(FALSE);
		if (temp[0] == '.')
		{
			temp_colour[i] = NONE;
			temp_kind[i] = INVALID;
		}
		else
		{
			if (temp[0] == 'w')
				temp_colour[i] = WHITE_COLOUR;
			else if (temp[0] == 'b')
				temp_colour[i] = BLACK_COLOUR;
			else
				FATAL_ERROR("invalid colour")
			if (temp[1] == 'O')
				temp_kind[i] = MAN;
			else if (temp[1] == 'X')
				temp_kind[i] = CROWN;
			else
				FATAL_ERROR("invalid kind")
		}
	}
	
	for (i = 0; i < BOARD_MAX; i++)
	{
		board[i] = &side;
		asm_board[i] = 0;
	}
	for (i = 0; i < 50; i++)
		board[map[i]] = &empty;
	
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		tot_count[icolour] = 0;
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			p = list[icolour][ikind];
			for (i = 0; i < LIST_MAX; i++)
			{
				p->board_colour = icolour;
				p->board_kind = INVALID;
				p->board_value = value[ikind];
				p->board_where = INVALID;
				p->board_key = hash_key[icolour][ikind];
				p->board_dir = dir[icolour][ikind];
				p->board_count = piece_count[icolour] + ikind;
				p->board_pat = pat[icolour][ikind];
				p++;
			}
			p = list[icolour][ikind];
			piece_count[icolour][ikind] = 0;
			for (i = 0; i < 50; i++)
			{
				if ((temp_colour[i] == icolour) and (temp_kind[i] == ikind))
				{
					piece_count[icolour][ikind]++;
					p->board_kind = ikind;
					p->board_where = map[i];
					board[map[i]] = p;
					asm_board[map[i]] = p->board_pat;
					p++;
				}
			}
			tot_count[icolour] += piece_count[icolour][ikind];
		}
	}
	
	node = node_stack;
	node->move_base = move_stack;
	node->stack_base = stack;
	node->node_key = return_cur_key();
	node->node_total_key = node->node_key;
	node->node_mat[WHITE_COLOUR] = return_mat_score(WHITE_COLOUR);
	node->node_mat[BLACK_COLOUR] = return_mat_score(BLACK_COLOUR);
	node->node_reverse = FALSE;
	null_move = FALSE;
	nthreats = 0;
	my_colour = WHITE_COLOUR;
	your_colour = BLACK_COLOUR;
	if ((err = check_list()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("list error")
	}
	if ((err = check_asm_board()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("asm error")
	}
	return(TRUE);
}

int ncapt, ncaptx;
int icapt[LIST_MAX];
int iboard[LIST_MAX];
int idir[LIST_MAX];
int jdir[LIST_MAX];
board_t *pcapt[LIST_MAX];
int nmoves;
int ntempo;

int gen_moves(void);
int gen_moves(void)
{
	board_t *p;
	int i, jboard1, kboard1;
	int *t;
	
	icapt[0] = 0;
	ncaptx = 0;
	nmoves = 0;
	ntempo = 0;
	for (p = man_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		board[p->board_where] = &empty;
		for (i = 0; i < 4; i++)
		{
			if (board[jboard1 = p->board_where + p->board_dir[i]] == &empty)
			{
				if ((i < 2) and (ncaptx == 0))
				{
					*move_top++ = stack_top;
					*stack_top++ = p->board_where;
					*stack_top++ = jboard1;
					*stack_top++ = 0;
					nmoves++;
					if (my_colour == WHITE_COLOUR)
					{
						if ((board[jboard1 - 6]->board_colour != BLACK_COLOUR) and
						    (board[jboard1 - 5]->board_colour != BLACK_COLOUR)) ntempo++;
					}
					else
					{
						if ((board[jboard1 + 5]->board_colour != WHITE_COLOUR) and
						    (board[jboard1 + 6]->board_colour != WHITE_COLOUR)) ntempo++;
					}
				}
			}
			if (board[jboard1]->board_colour != your_colour) continue;
			if (board[kboard1 = jboard1 + p->board_dir[i]] == &empty)
			{
				idir[1] = p->board_dir[i];
				ncapt = 0;
				label_next_man_capt:
				pcapt[ncapt + 1] = board[jboard1];
				board[jboard1] = &side;
				icapt[ncapt + 1] = jboard1;
				iboard[ncapt + 1] = kboard1;
				ncapt++;
				if (ncapt > ncaptx)
				{
					ncaptx = ncapt;
					move_top = node->move_base;
					stack_top = node->stack_base;
					nmoves = 0;
					ntempo = 0;
				}
				if (ncapt == ncaptx)
				{
					*move_top++ = stack_top;
					*stack_top++ = p->board_where;
					*stack_top++ = kboard1;
					t = icapt + ncapt;
					while((*stack_top++ = *t--) != 0);
					nmoves++;
					if (my_colour == WHITE_COLOUR)
					{
						if ((board[kboard1 - 6]->board_colour != BLACK_COLOUR) and
						    (board[kboard1 - 5]->board_colour != BLACK_COLOUR)) ntempo++;
					}
					else
					{
						if ((board[kboard1 + 5]->board_colour != WHITE_COLOUR) and
						    (board[kboard1 + 6]->board_colour != WHITE_COLOUR)) ntempo++;
					}
				}
				jdir[ncapt] = 11 - ABS(idir[ncapt]);
				if (board[jboard1 = iboard[ncapt] + jdir[ncapt]]->board_colour == your_colour)
				{
					if (board[kboard1 = jboard1 + jdir[ncapt]] == &empty)
					{
						idir[ncapt + 1] = jdir[ncapt];
						goto label_next_man_capt;
					}
				}
				label_neg_man_dir:
				jdir[ncapt] = -jdir[ncapt];
				if (board[jboard1 = iboard[ncapt] + jdir[ncapt]]->board_colour == your_colour)
				{
					if (board[kboard1 = jboard1 + jdir[ncapt]] == &empty)
					{
						idir[ncapt + 1] = jdir[ncapt];
						goto label_next_man_capt;
					}
				}
				label_man_dir:
				jdir[ncapt] = 0;
				if (board[jboard1 = iboard[ncapt] + idir[ncapt]]->board_colour == your_colour)
				{
					if (board[kboard1 = jboard1 + idir[ncapt]] == &empty)
					{
						idir[ncapt + 1] = idir[ncapt];
						goto label_next_man_capt;
					}
				}
				label_undo_man_capt:
				board[icapt[ncapt]] = pcapt[ncapt];
				ncapt--;
				if (ncapt > 0)
				{
					if (jdir[ncapt] > 0) goto label_neg_man_dir;
					if (jdir[ncapt] < 0) goto label_man_dir;
					goto label_undo_man_capt;
				}
			}
		}
		board[p->board_where] = p;
	}
	for (p = crown_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		board[p->board_where] = &empty;
		for (i = 0; i < 4; i++)
		{
			jboard1 = p->board_where;
			while(board[jboard1 += p->board_dir[i]] == &empty)
			{
				if (ncaptx == 0)
				{
					*move_top++ = stack_top;
					*stack_top++ = p->board_where;
					*stack_top++ = jboard1;
					*stack_top++ = 0;
					nmoves++;
				}
			}
			if (board[jboard1]->board_colour != your_colour) continue;
			if (board[kboard1 = jboard1 + p->board_dir[i]] == &empty)
			{
				idir[1] = p->board_dir[i];
				ncapt = 0;
				label_next_crown_capt:
				pcapt[ncapt + 1] = board[jboard1];
				board[jboard1] = &side;
				icapt[ncapt + 1] = jboard1;
				iboard[ncapt + 1] = kboard1;
				ncapt++;
				label_patch:
				if (ncapt > ncaptx)
				{
					ncaptx = ncapt;
					move_top = node->move_base;
					stack_top = node->stack_base;
					nmoves = 0;
					ntempo = 0;
				}
				if (ncapt == ncaptx)
				{
					*move_top++ = stack_top;
					*stack_top++ = p->board_where;
					*stack_top++ = kboard1;
					t = icapt + ncapt;
					while((*stack_top++ = *t--) != 0);
					nmoves++;
				}
				jdir[ncapt] = 11 - ABS(idir[ncapt]);
				jboard1 = iboard[ncapt];
				while(board[jboard1 += jdir[ncapt]] == &empty);
				if (board[jboard1]->board_colour == your_colour)
				{
					if (board[kboard1 = jboard1 + jdir[ncapt]] == &empty)
					{
						idir[ncapt + 1] = jdir[ncapt];
						goto label_next_crown_capt;
					}
				}
				label_neg_crown_dir:
				jdir[ncapt] = -jdir[ncapt];
				jboard1 = iboard[ncapt];
				while(board[jboard1 += jdir[ncapt]] == &empty);
				if (board[jboard1]->board_colour == your_colour)
				{
					if (board[kboard1 = jboard1 + jdir[ncapt]] == &empty)
					{
						idir[ncapt + 1] = jdir[ncapt];
						goto label_next_crown_capt;
					}
				}
				label_crown_dir:
				jdir[ncapt] = 0;
				if (board[jboard1 = iboard[ncapt] + idir[ncapt]] == &empty)
				{
					kboard1 = iboard[ncapt] = jboard1;
					goto label_patch;
				}
				if (board[jboard1]->board_colour == your_colour)
				{
					if (board[kboard1 = jboard1 + idir[ncapt]] == &empty)
					{
						idir[ncapt + 1] = idir[ncapt];
						goto label_next_crown_capt;
					}
				}
				label_undo_crown_capt:
				board[icapt[ncapt]] = pcapt[ncapt];
				ncapt--;
				if (ncapt > 0)
				{
					if (jdir[ncapt] > 0) goto label_neg_crown_dir;
					if (jdir[ncapt] < 0) goto label_crown_dir;
					goto label_undo_crown_capt;
				}
			}
		}
		board[p->board_where] = p;
	}
	return(nmoves);
}

int gen_inverse_moves(void);
int gen_inverse_moves(void)
{
	board_t *p;
	int i, jboard1;
	
	icapt[0] = 0;
	ncaptx = 0;
	nmoves = 0;
	ntempo = 0;
	for (p = man_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		for (i = 2; i < 4; i++)
		{
			if (board[jboard1 = p->board_where + p->board_dir[i]] == &empty)
			{
				*move_top++ = stack_top;
				*stack_top++ = p->board_where;
				*stack_top++ = jboard1;
				*stack_top++ = 0;
				nmoves++;
			}
		}
	}
	for (p = crown_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		for (i = 0; i < 4; i++)
		{
			jboard1 = p->board_where;
			while(board[jboard1 += p->board_dir[i]] == &empty)
			{
				*move_top++ = stack_top;
				*stack_top++ = p->board_where;
				*stack_top++ = jboard1;
				*stack_top++ = 0;
				nmoves++;
			}
		}
	}
	return(nmoves);
}

int can_capture(int, int);
int can_capture(int my_colour, int your_colour)
{
	board_t *p;
	int i, jboard1;

	for (p = man_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		for (i = 0; i < 4; i++)
		{
			jboard1 = p->board_where + p->board_dir[i];
			if (board[jboard1]->board_colour != your_colour) continue;
			if (board[jboard1 + p->board_dir[i]] == &empty)
				return(TRUE);
		}
	}
	for (p = crown_list[my_colour]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		for (i = 0; i < 4; i++)
		{
			jboard1 = p->board_where;
			while(board[jboard1 += p->board_dir[i]] == &empty);
			if (board[jboard1]->board_colour != your_colour) continue;
			if (board[jboard1 + p->board_dir[i]] == &empty)
				return(TRUE);
		}
	}
	return(FALSE);
}

void do_move(move_t *m)
{
	board_t *p;
	int from, to;
#if DEBUG_LEVEL > 0
	int err;
	hash_key_t temp_key;
#endif
	
	node->node_move = m;
	node->capt_base = capt_top;

	node[1].node_key = node->node_key;
	node[1].node_mat[WHITE_COLOUR] = node->node_mat[WHITE_COLOUR];
	node[1].node_mat[BLACK_COLOUR] = node->node_mat[BLACK_COLOUR];
	node++;
	inode++;
	
	node->node_reverse = FALSE;
	if (m == NULL) goto label_null;
	
	from = *m++;
	to = *m++;
	BUG(row[from] == INVALID)
	BUG(row[to] == INVALID)
	
	p = board[from];
	UPDATE_KEY(node->node_key, p->board_key[from])
	board[from] = &empty;
	asm_board[from] = 0;
	if ((p->board_kind == MAN) and (row[to] == prom_row[my_colour]))
	{
		node->prom = p;
		p->board_where = INVALID;
		node->node_mat[my_colour] -= p->board_value;
		(*p->board_count)--;
		for (p = crown_list[my_colour]; p->board_kind != INVALID; p++);
		p->board_kind = CROWN;
		p->board_where = to;
		board[to] = p;
		asm_board[to] = p->board_pat;
		UPDATE_KEY(node->node_key, p->board_key[to])
		node->node_mat[my_colour] += p->board_value;
		(*p->board_count)++;
	}
	else
	{
		node->prom = NULL;
		UPDATE_KEY(node->node_key, p->board_key[to])
		p->board_where = to;
		board[to] = p;
		asm_board[to] = p->board_pat;
	}
	
	while((from = *m++) != 0)
	{
		p = board[from];
		BUG(p->board_colour != your_colour)
		*capt_top++ = p;
		p->board_where = INVALID;
		board[from] = &empty;
		asm_board[from] = 0;
		UPDATE_KEY(node->node_key, p->board_key[from])
		node->node_mat[your_colour] -= p->board_value;
		tot_count[your_colour]--;
		(*p->board_count)--;
	}

	label_null:
	node->node_total_key = node->node_key;
	update_key(&(node->node_total_key));
#if DEBUG_LEVEL > 0
	if ((err = check_list()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		display_move("", node[-1].node_move, "\n");
		display_board();
		display_list();
		FATAL_ERROR("do list error")
	}
	if ((err = check_asm_board()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("do asm error")
	}
	temp_key = return_cur_key();
	if ((temp_key.lo != node->node_key.lo) or
	    (temp_key.hi != node->node_key.hi))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		display_move("", node[-1].node_move, "\n");
		display_board();
		display_list();
		FATAL_ERROR("do key error")
	}
	if (node->node_mat[WHITE_COLOUR] != return_mat_score(WHITE_COLOUR))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		display_move("", node[-1].node_move, "\n");
		display_board();
		display_list();
		FATAL_ERROR("do white mat error")
	}
	if (node->node_mat[BLACK_COLOUR] != return_mat_score(BLACK_COLOUR))
	{
		my_printf("%s\n", colour_as_string[my_colour]);
		display_move("", node[-1].node_move, "\n");
		display_board();
		display_list();
		FATAL_ERROR("do black mat error")
	}
#endif
	if (my_colour == WHITE_COLOUR)
	{
		my_colour = BLACK_COLOUR;
		your_colour = WHITE_COLOUR;
	}
	else
	{
		my_colour = WHITE_COLOUR;
		your_colour = BLACK_COLOUR;
	}
#undef DO_CAPT
#undef DO_MOVE
}

void undo_move(void)
{
	move_t *m;
	board_t *p;
	int from, to;
#if DEBUG_LEVEL > 0
	int err;
#endif

	if (my_colour == WHITE_COLOUR)
	{
		my_colour = BLACK_COLOUR;
		your_colour = WHITE_COLOUR;
	}
	else
	{
		my_colour = WHITE_COLOUR;
		your_colour = BLACK_COLOUR;
	}
	capt_top = node[-1].capt_base;
	m = node[-1].node_move;
	if (m == NULL) goto label_null;
	from = *m++;
	to = *m++;
	BUG(row[from] == INVALID)
	BUG(row[to] == INVALID)
	p = board[to];
	board[to] = &empty;
	asm_board[to] = 0;
	if (node->prom != NULL)
	{
		p->board_kind = p->board_where = INVALID;
		(*p->board_count)--;
		p = node->prom;
		p->board_where = from;
		board[from] = p;
		asm_board[from] = p->board_pat;
		(*p->board_count)++;
	}
	else
	{
		p->board_where = from;
		board[from] = p;
		asm_board[from] = p->board_pat;
	}
	
	while((from = *m++) != 0)
	{
		p = *capt_top++;
		p->board_where = from;
		board[from] = p;
		asm_board[from] = p->board_pat;
		tot_count[your_colour]++;
		(*p->board_count)++;
	}

	label_null:
	node--;
	inode--;
	BUG(inode < 0)
	capt_top = node->capt_base;
#if DEBUG_LEVEL > 0
	if ((err = check_list()) > 0)
	{
		my_printf("%s err=%d\n", colour_as_string[my_colour], err);
		display_move("", node->node_move, "\n");
		display_board();
		display_list();
		FATAL_ERROR("undo list error")
	}
	if ((err = check_asm_board()) > 0)
	{
		my_printf("err=%d\n", err);
		display_board();
		display_list();
		FATAL_ERROR("undo asm error")
	}
#endif
}

void heapsort(int n, int *);
void heapsort(int n, int *a)
{
	int l, r, b, i, j;
	
	if (n < 2) return;
	l = n / 2 + 1;
	r = n;
	h2:
	if (l > 1)
	{
		l--;
		b = a[l];
	}
	else
	{
		b = a[r];
		a[r] = a[1];
		r--;
		if (r == 1)
		{
			a[1] = b;
			goto h9;
		}		
	}
	j = l;
	h4:
	i = j;
	j = 2 * j;
	if (j == r) goto h6;
	if (j > r) goto h8;
	if (a[j] < a[j + 1]) j++;
	h6:
	if (b >= a[j]) goto h8;
	a[i] = a[j];
	goto h4;
	h8:
	a[i] = b;
	goto h2;
	h9:
	for (i = 1; i < n; i++)
		BUG(a[i] > a[i + 1])
}

#define ENDGAME_BITS 7
#define ENDGAME_SIZE BIT(ENDGAME_BITS)
#define ENDGAME_MASK (~(~0 << ENDGAME_BITS))
#define ENDGAME_MAX  4
endgame_t only_dbase = 0;
#if MACHINE == DOS
endgame_t *endgame_pointer;
uint_t endgame_dbase[ENDGAME_SIZE];
#else
endgame_t *endgame_pointer;
endgame_t *endgame_dbase[ENDGAME_SIZE];
#endif
long endgame_size[ENDGAME_SIZE];
int wc_key[ENDGAME_MAX], wm_key[ENDGAME_MAX], bc_key[ENDGAME_MAX], bm_key[ENDGAME_MAX];

void set_endgame_pointer(int, int);
void set_endgame_pointer(int my_colour, int your_colour)
{
#define MIR(X) (49 - (X))

	board_t *p;
	int temp[20];
	int ntemp, norg;
	int i;
	long j;
	int m;
	int icolour, ikind;
	
	DUMMY("%d", your_colour)
#if MACHINE == MAC
	BUG(Button())
#endif
	if (WTOT == 0)
	{
		if (my_colour == WHITE_COLOUR)
			endgame_pointer = &only_dbase;
		else
			endgame_pointer = NULL;
		return;
	}
	else if (BTOT == 0)
	{
		if (my_colour == WHITE_COLOUR)
			endgame_pointer = NULL;
		else
			endgame_pointer = &only_dbase;
		return;
	}
	
	if ((WTOT + BTOT) > 3)
	{
		endgame_pointer = NULL;
		return;
	}
	if (my_colour == WHITE_COLOUR)
	{
		m = 0;
		m ^= wc_key[piece_count[WHITE_COLOUR][CROWN]];
		m ^= wm_key[piece_count[WHITE_COLOUR][MAN]];
		m ^= bc_key[piece_count[BLACK_COLOUR][CROWN]];
		m ^= bm_key[piece_count[BLACK_COLOUR][MAN]];
		if (endgame_dbase[m] == NULL)
		{
			endgame_pointer = NULL;
			return;
		}
		ntemp = 0;
		for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
		{
			for (ikind = MAN; ikind <= CROWN; ikind++)
			{
				norg = ntemp;
				for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
				{
					if (p->board_where == INVALID) continue;
					temp[ntemp++] = inverse_map[p->board_where];
				}
				heapsort(ntemp - norg, temp + norg - 1);
			}
		}
	}
	else
	{
		m = 0;
		m ^= wc_key[piece_count[BLACK_COLOUR][CROWN]];
		m ^= wm_key[piece_count[BLACK_COLOUR][MAN]];
		m ^= bc_key[piece_count[WHITE_COLOUR][CROWN]];
		m ^= bm_key[piece_count[WHITE_COLOUR][MAN]];
		if (endgame_dbase[m] == NULL)
		{
			endgame_pointer = NULL;
			return;
		}
		ntemp = 0;
		for (icolour = BLACK_COLOUR; icolour >= WHITE_COLOUR; icolour--)
		{
			for (ikind = MAN; ikind <= CROWN; ikind++)
			{
				norg = ntemp;
				for (p = list[icolour][ikind]; p->board_kind != INVALID; p++)
				{
					if (p->board_where == INVALID) continue;
					temp[ntemp++] = MIR(inverse_map[p->board_where]);
				}
				heapsort(ntemp - norg, temp + norg - 1);
			}
		}
	}
	j = temp[0];
	for (i = 1; i < ntemp; i++)
		j = j * 50L + temp[i];
#if MACHINE == DOS
	XMShandle = endgame_dbase[m];
	XMSoffset = j;
	endgame_pointer = &only_dbase;
#else
	endgame_pointer = endgame_dbase[m] + j;
#endif
}

endgame_t get_endgame(void);
endgame_t get_endgame(void)
{
#if MACHINE == DOS
	x2c.srchandle = XMShandle;
	x2c.srcoffset = XMSoffset;
	XMS_move(&x2c);
	return (XMSresult[0]);
#else
	return(*endgame_pointer);
#endif
}

void put_endgame(endgame_t);
void put_endgame(endgame_t what)
{
#if MACHINE == DOS
	get_endgame();
	XMSresult[0] = result;
	c2x.dsthandle = XMShandle;
	c2x.dstoffset = XMSoffset;
	XMS_move(&c2x);
#else
	*endgame_pointer = what;
#endif
}

int return_endgame_score(int my_colour, int your_colour)
{
	int temp;
	
	set_endgame_pointer(my_colour, your_colour);
	if (endgame_pointer == NULL)
		return(SCORE_INVAL);
	temp = get_endgame();
	if (temp == ENDGAME_DRAW)
		return(0);
	else
	{
		if (temp > 0)
			return(SCORE_MATE - inode - temp);
		else
			return(-SCORE_MATE + inode - temp);
	}
}

void set_score(double *y)
{
	DUMMY("%f", *y)
}

int set_pat_board(void);
int set_pat_board(void)
{
	board_t *p;
	int sum;
	
	memcpy(b, bempty, 51);
	sum = 0;
	for (p = list[WHITE_COLOUR][MAN]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		b[pat_map[p->board_where]] = p->board_pat;
		sum += white_man_visit[p->board_where];
	}
	for (p = list[WHITE_COLOUR][CROWN]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		b[pat_map[p->board_where]] = p->board_pat;
	}
	for (p = list[BLACK_COLOUR][MAN]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		b[pat_map[p->board_where]] = p->board_pat;
		sum -= black_man_visit[p->board_where];
	}
	for (p = list[BLACK_COLOUR][CROWN]; p->board_kind != INVALID; p++)
	{
		if (p->board_where == INVALID) continue;
		b[pat_map[p->board_where]] = p->board_pat;
	}
	return(sum);
}

void mirror_pat_board(void);
void mirror_pat_board(void)
{
	int i;
	pat_t p, q;
	
	for (i = 1; i <= 25; i++)
	{
		if (WM(i)) p = black_pat[MAN];
		else if (BM(i)) p = white_pat[MAN];
		else if (WC(i)) p = black_pat[CROWN];
		else if (BC(i)) p = white_pat[CROWN];
		else p = 0;
		if (WM(51 - i)) q = black_pat[MAN];
		else if (BM(51 - i)) q = white_pat[MAN];
		else if (WC(51 - i)) q = black_pat[CROWN];
		else if (BC(51 - i)) q = white_pat[CROWN];
		else q = 0;
		b[i] = q;
		b[51 - i] = p;
	}
}

void display_score(void)
{
	int nwhite_mat, nblack_mat;
	int nwhite_pos, nblack_pos;
	int nwhite_tot, nblack_tot;
	int ndiff, nendgame;
	
	nendgame = return_endgame_score(WHITE_COLOUR, BLACK_COLOUR);
	my_printf("endgame=%d\n", nendgame);
	
	nwhite_mat = return_mat_score(WHITE_COLOUR);
	nblack_mat = return_mat_score(BLACK_COLOUR);
	set_pat_board();
	nwhite_pos = return_pat_score();
	mirror_pat_board();
	nblack_pos = return_pat_score();
	nwhite_tot = nwhite_pos;		
	nblack_tot = nblack_pos;

	my_printf("mat score white=%d black=%d\n", nwhite_mat, nblack_mat);
	my_printf("pos score white=%d black=%d\n", nwhite_pos, nblack_pos);
	my_printf("tot pos score white=%d black=%d\n", nwhite_tot, nblack_tot);
	ndiff = nwhite_mat - nblack_mat + nwhite_tot - nblack_tot;
	my_printf("diff score=%d\n", ndiff);
}

int draw_by_repetition(int strict)
{
	node_t *temp_node;
	int ncount;
	
  temp_node = node;
  ncount = 1;
  while(TRUE)
  {
  	if (!temp_node->node_reverse) break;
  	temp_node--;
  	if (!temp_node->node_reverse) break;
  	temp_node--;
  	if ((node->node_total_key.lo == temp_node->node_total_key.lo) and
  	    (node->node_total_key.hi == temp_node->node_total_key.hi))
  	{
  		ncount++;
  		if (!strict)
  			return(TRUE);
  		else if (ncount > 2)
  			return(TRUE);
  	}
  }
  return(FALSE);
}

/*
heapsort routine. Sorts the move strength a and the moves m.

void heapsort(int n, int *, move_t *);
void heapsort(int n, int *a, move_t *m)
{
	int l, r, b, i, j;
	move_t t;
	
	if (n < 2) return;
	l = n / 2 + 1;
	r = n;
	h2:
	if (l > 1)
	{
		l--;
		b = a[l]; t = m[l];
	}
	else
	{
		b = a[r]; t = m[r];
		a[r] = a[1]; m[r] = m[1];
		r--;
		if (r == 1)
		{
			a[1] = b; m[1] = t;
			goto h9;
		}		
	}
	j = l;
	h4:
	i = j;
	j = 2 * j;
	if (j == r) goto h6;
	if (j > r) goto h8;
	if (a[j] < a[j + 1]) j++;
	h6:
	if (b >= a[j]) goto h8;
	a[i] = a[j]; m[i] = m[j];
	goto h4;
	h8:
	a[i] = b; m[i] = t;
	goto h2;
	h9:
	for (i = 1; i < n; i++)
		BUG(a[i] > a[i + 1])
}

*/

#define LOG_NULL_MOVE

score_t alpha_beta(score_t, score_t, int, var_t *);
score_t alpha_beta(score_t alpha, score_t beta, int depth, var_t *var_base)
{
#define RETURN(X)\
	{best_score = (X); goto label_neg;}
	
	score_t best_score, temp_score;
#if MACHINE == DOS
	cache_t huge *p;
#else
	cache_t *p;
#endif
	int best_move;
	int nmoves;
	int imove, jmove;
	int depth_next;
	score_t threat_score = SCORE_INVAL;
	int mat_diff = node->node_mat[my_colour] - node->node_mat[your_colour];
#ifdef LOG_NULL_MOVE
	int log_null = FALSE;
#endif
	var_t *var, *var_top;
	
#if DEBUG_LEVEL > 0
	node_t *node_debug = node;
#endif
#if MACHINE == MAC
	if (Button()) return(SCORE_INVAL);
#endif
	
	BUG(alpha >= beta)
	BUG(depth < 0)
	
	var_top = var_base;
	*var_top++ = INVALID;
	
	clock_end = clock();
	if (clock_end < clock_begin)
	{
		my_printf("clock time cycle detected!\n");
		clock_begin = clock_end;
	}
	else
	{
		if ((clock_end - clock_begin) > (CLOCKS_PER_SEC / 10))
		{
			time_used += (double) (clock_end - clock_begin) / CLOCKS_PER_SEC;
			clock_begin = clock_end;
#ifdef GUI
			if (check_time())
				return (SCORE_INVAL);
#else
			if (time_used > TimeLimit) return(SCORE_INVAL);
#endif
		}
	}
	
	temp_score = alpha;
	alpha = -beta;
	beta = -temp_score;
	
	node->move_base = move_top;
	node->stack_base = stack_top;

	best_score = -SCORE_LIMIT;
	best_move = INVALID;
	
	p = cache[my_colour] + (node->node_total_key.lo & CACHE_MASK);
	if (!(p->cache_key.lo & CACHE_INVAL) and
	    (p->cache_key.hi == node->node_total_key.hi) and
	    ((p->cache_key.lo & ~CACHE_KIND_MASK) == (node->node_total_key.lo & ~CACHE_KIND_MASK)))
	{
		best_move = p->cache_best;
		if ((p->cache_key.lo & CACHE_DEPTH_MASK) >= depth)
		{
			if (p->cache_key.lo & CACHE_TRUE)
				RETURN(p->cache_score)
			else if (p->cache_key.lo & CACHE_LOWER)
			{
				if (p->cache_score >= beta)
					RETURN(p->cache_score)
				else if (p->cache_score > alpha)
					alpha = p->cache_score;
			}
			else if (p->cache_key.lo & CACHE_UPPER)
			{
				if (p->cache_score <= alpha)
					RETURN(p->cache_score)
				else if (p->cache_score < beta)
					beta = p->cache_score;
			}
		}
		else
			p->cache_key.lo &= ~CACHE_TRUE;
	}
	
	if (draw_by_repetition(FALSE)) RETURN(0)
	
	if (UseEndgame)
	{
		temp_score = return_endgame_score(my_colour, your_colour);
		if (temp_score != SCORE_INVAL)
		{
			/*
			display_board();
			my_printf("%s endgame %d\n", colour_as_string[my_colour], temp_score);
			*/
			best_score = temp_score;
			goto label_neg;
		}
	}

	nmoves = gen_moves();
	if (nmoves == 0)
		RETURN(-SCORE_MATE + inode)
#if DEBUG_LEVEL > 0
	if ((ncaptx > 0) and !can_capture(my_colour, your_colour))
	{
		display_board();
		my_printf("%s\n", colour_as_string[my_colour]);
		FATAL_ERROR("capture error")
	}
#endif

	if ((depth == 0) and (ncaptx == 0) and (nmoves > 2))
	{
#if MACHINE == DOS
		score_cache_t huge *p;
#else
		score_cache_t *p;
#endif
		
		nevals_tot++;
		best_score = node->node_mat[my_colour] - node->node_mat[your_colour];
		if (UsePos == 0) goto label_neg;
		
		p = score_cache[my_colour] + (node->node_total_key.lo & SCORE_CACHE_MASK);
		if ((p->cache_score != SCORE_INVAL) and
		    (p->cache_key.lo == node->node_total_key.lo) and
		    (p->cache_key.hi == node->node_total_key.hi))
		{
			best_score += p->cache_score;
			nhits++;
		}
		else
		{
#if DEBUG_LEVEL > 0
			pat_t bug[51];
#endif
			temp_score = set_pat_board();
#if DEBUG_LEVEL > 0
			memcpy(bug + 1, b + 1, 50);
#endif
			temp_score += return_pat_score();
			mirror_pat_board();
			temp_score -= return_pat_score();
#if DEBUG_LEVEL > 0
			mirror_pat_board();
			BUG(memcmp(bug + 1, b + 1, 50) != 0)
#endif
			if (my_colour == BLACK_COLOUR)
				temp_score = -temp_score;
			best_score += temp_score;
			nevals++;
			p->cache_score = temp_score;
			p->cache_key = node->node_total_key;
		}
		goto label_neg;
	}
	
	if (ncaptx > 0)
		depth_next = depth;
	else if (nmoves > 4)
	{
		depth_next = depth - 1;

#ifdef NULL_MOVE
		if (!principle_variation and !null_move and
		    (depth >= NULL_MOVE_DEPTH) and (piece_count[my_colour][MAN] > 5))
		{
			null_move = TRUE;
			do_move(NULL);
			temp_score = alpha_beta(beta - 1, beta, depth_next - NULL_MOVE_REDUC, var_top);
			undo_move();
			null_move = FALSE;
			if (temp_score == SCORE_INVAL) return(SCORE_INVAL);
#ifndef LOG_NULL_MOVE
			if (temp_score >= beta) RETURN(temp_score)
#else
			if (temp_score >= beta) log_null = TRUE;
#endif
		}
#endif

#ifdef THREAT	
		if (!null_move and
		    (nthreats < NTHREAT_MAX) and (depth >= THREAT_DEPTH))
		{
			null_move = TRUE;
			do_move(NULL);
			threat_score = alpha_beta(mat_diff - THREAT_MARGIN - 1, mat_diff - THREAT_MARGIN,
				depth_next - THREAT_REDUC, var_top);
			undo_move();
			null_move = FALSE;
			if (threat_score == SCORE_INVAL) return(SCORE_INVAL);
			if ((threat_score + THREAT_MARGIN) >= mat_diff) threat_score = SCORE_INVAL;
		}
#endif
	}
	else if (nmoves > 2)
		depth_next = depth - 1;
	else
		depth_next = depth;
	
	ngens++;
	if (best_move == INVALID)
	{
		best_move = 0;
	}
	else if (best_move >= nmoves)
	{
		my_printf("hash fault!\n");
		best_move = 0;
	}

	nodes_tot++;
	for (imove = INVALID; imove < nmoves; imove++)
	{
		if (imove == INVALID)
			jmove = best_move;
		else
			jmove = imove;
		if (node->move_base[jmove] == NULL) continue;
		do_move(node->move_base[jmove]);
		ndos++;
		if (threat_score == SCORE_INVAL)
			temp_score = alpha_beta(MAX(alpha, best_score), beta, depth_next, var_top);
		else
		{
			temp_score = alpha_beta(mat_diff - THREAT_MARGIN - 1, mat_diff - THREAT_MARGIN,
				depth_next - THREAT_REDUC, var_top);
			if (temp_score == SCORE_INVAL) return(SCORE_INVAL);
			if ((temp_score + THREAT_MARGIN) < mat_diff)
			{
				temp_score = alpha_beta(MAX(alpha, best_score), beta, depth_next, var_top);
			}
			else
			{
				nthreats++;
				temp_score = alpha_beta(MAX(alpha, best_score), beta, depth_next + 1, var_top);
				nthreats--;
			}
		}
		undo_move();
		node->move_base[jmove] = NULL;
		if (temp_score == SCORE_INVAL) return(SCORE_INVAL);
		if (temp_score > best_score)
		{
			best_score = temp_score;
			best_move = jmove;
			var = var_base;
			*var++ = jmove;
			while ((*var++ = *var_top++) != INVALID);
			var_top = var;
			var = var_base;
			if (best_score >= beta) goto label_store;
		}
	}
#ifdef LOG_NULL_MOVE
	if (log_null)
	{
		display_board();
		my_printf("%snull move failure!\n", colour_as_string[my_colour]);
		nmoves = gen_moves();
		my_printf("nmoves=%d ntempo=%d\n", nmoves, ntempo);
	}
#endif
	label_store:
	
	if (depth > CACHE_DEPTH_MASK) goto label_neg;
	p->cache_key.hi = node->node_total_key.hi;
	p->cache_key.lo = ((node->node_total_key.lo & ~CACHE_KIND_MASK) | depth);
	p->cache_score = best_score;
	if (best_score <= alpha)
		p->cache_key.lo |= CACHE_UPPER;
	else if (best_score >= beta)
		p->cache_key.lo |= CACHE_LOWER;
	else
		p->cache_key.lo |= CACHE_TRUE;
	p->cache_best = best_move;
	
	label_neg:
	BUG(best_score == -SCORE_LIMIT)
#if DEBUG_LEVEL > 0
	BUG(node != node_debug)
#endif
	move_top = node->move_base;
	stack_top = node->stack_base;
	return(-best_score);
#undef RETURN
}

int nroot;
root_t *root;
int jroot;

void gen_root(void)
{
	int i;
				
	nroot = 0;
	move_top = move_stack;
	stack_top = stack;
	capt_top = capt_stack;
	node->move_base = move_top;
	node->stack_base = stack_top;
	nroot = gen_moves();
	for (i = 0; i < nroot; i++)
	{
		root[i].root_move = i;
		root[i].root_score = 0;
		root[i].root_hit = 0;
	}
	jroot = 0;
}

int search_root(char *move, int verbose)
{
	int iroot, temp, nhit;
	size_t len;
	
	len = strlen(move);
	nhit = 0;
	temp = INVALID;
	for (iroot = 0; iroot < nroot; iroot++)
	{
		move_to_string(node->move_base[root[iroot].root_move], TRUE);
		if (strncmp(move_string, move, len) == 0) 
		{
			nhit++;
			temp = iroot;
		}
	}
	if (nhit == 1) return(temp);
	if (verbose)
	{
		display_board();
		my_printf("move=%s\n", move);
		for (iroot = 0; iroot < nroot; iroot++)
		{
			move_to_string(node->move_base[root[iroot].root_move], TRUE);
			my_printf("%s\n", move_string);
		}
	}
	return(INVALID);
}

void analyze_cache(int, var_t *);
void analyze_cache(int best_move, var_t *var_base)
{
	int nmoves;
	
	BUG(best_move < 0)
	move_to_string(node->move_base[best_move], FALSE);
	my_printf(" %s", move_string);
#ifdef GUI
	print_cache_move(node->move_base[best_move]);
#endif
	if (draw_by_repetition(FALSE))
	{
		my_printf(" rep!");
		return;
	}
	if (*var_base != INVALID)
	{
		do_move(node->move_base[best_move]);
		node->move_base = move_top;
		node->stack_base = stack_top;
		nmoves = gen_moves();
		if (*var_base >= nmoves)
		{
			my_printf("\n%d %d\n", *var_base, nmoves);
			FATAL_ERROR("xx")
		}
		analyze_cache(*var_base, var_base + 1);
		move_top = node->move_base;
		stack_top = node->stack_base;
		undo_move();
	}
}

void clear_cache(void);
void clear_cache(void)
{
	long i;
	int icolour;

	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (i = 0; i < (CACHE_MASK + 1); i++)
			cache[icolour][i].cache_key.lo = CACHE_INVAL;
		for (i = 0; i < (SCORE_CACHE_MASK + 1); i++)
			score_cache[icolour][i].cache_score = SCORE_INVAL;
	}
}

void sort_root(void)
{
	int iroot, kroot;
	root_t temp;
	
	clear_cache();
	inode = 0;
	
	clock_begin = clock();
	time_used = 0;
	TimeLimit = 100;
	
	ngens = ndos = 0;
	nevals = nhits = nextends = 0;

	principle_variation = TRUE;
	for (iroot = 0; iroot < nroot; iroot++)
	{
		do_move(node->move_base[root[iroot].root_move]);
		root[iroot].root_score = alpha_beta(-SCORE_LIMIT, SCORE_LIMIT, 0, var_stack);
		undo_move();
		my_printf("ts");
		analyze_cache(root[iroot].root_move, var_stack);
		my_printf(" %.3f\n", (double) root[iroot].root_score);
	}
	principle_variation = FALSE;
	BUG(null_move != FALSE)
	BUG(nthreats != 0)
	my_printf("time=%.2f stats=%ld/%ld/%ld/%ld/%ld\n",
		time_used, ngens, ndos, nevals, nhits, nextends);
	for (iroot = 0; iroot < (nroot - 1); iroot++)
	{
		jroot = iroot;
		for (kroot = iroot + 1; kroot < nroot; kroot++)
			if (root[kroot].root_score > root[jroot].root_score) jroot = kroot;
		if (jroot != iroot)
		{
			temp = root[iroot];
			root[iroot] = root[jroot];
			root[jroot] = temp;
		}
	}
	jroot = 0;
}

#define MARGIN (MAN_SCORE / 2)
score_t begin_score;

void do_root(void)
{
	int depth;
	score_t temp_score, bound_score, fh_score;
	int depth_next;
	score_t alpha, beta;
	int iroot;
#ifndef GUI
	FILE *f;
#endif
		
	clear_cache();

	jroot = 0;
	begin_score = node->node_mat[my_colour] - node->node_mat[your_colour];
	root[0].root_score = begin_score;
	inode = 0;
	
	clock_begin = clock();
	time_used = 0;
	
	iroot = 0;
	temp_score = SCORE_INVAL;
	
	nodes_tot = nevals_tot = 0;
	for (depth = 1; depth <= DepthMax; depth++)
	{
		depth_next = depth - 1;
		ngens = ndos = 0;
		nevals = nhits = nextends = 0;
		
		alpha = root[jroot].root_score - MARGIN;
		beta = root[jroot].root_score + MARGIN;
		
		principle_variation = TRUE;
		do_move(node->move_base[root[jroot].root_move]);
		temp_score = alpha_beta(alpha, beta, depth_next, var_stack);
		undo_move();
		if (temp_score == SCORE_INVAL) goto label_final;
		
		do_move(node->move_base[root[jroot].root_move]);
		if (temp_score <= alpha)
		{
			bound_score = temp_score;
			temp_score = alpha_beta(-SCORE_LIMIT, bound_score, depth_next, var_stack);
			if (temp_score == SCORE_INVAL)
			{
				undo_move();
				goto label_final;
			}
			if (temp_score > bound_score) my_printf("research error!\n");
		}
		else if (temp_score >= beta)
		{
			bound_score = temp_score;
			temp_score = alpha_beta(bound_score, SCORE_LIMIT, depth_next, var_stack);
			if (temp_score == SCORE_INVAL)
			{
				undo_move();
				goto label_final;
			}
			if (temp_score < bound_score) my_printf("research error!\n");
		}
		undo_move();
		root[jroot].root_score = bound_score = temp_score;
		my_printf("pv %d", depth);
#ifdef GUI
		print_cache(root[jroot].root_move, var_stack);
#else
		analyze_cache(root[jroot].root_move, var_stack);
#endif
		my_printf(" %.3f\n", (double) root[jroot].root_score);
#ifdef GUI
		print_best_move(depth, jroot);
#endif

		principle_variation = FALSE;
		for (iroot = 0; iroot < nroot; iroot++)
		{
			if (iroot == jroot) continue;
#ifdef GUI
			print_cur_move(depth, iroot);
#endif
			do_move(node->move_base[root[iroot].root_move]);
			temp_score = alpha_beta(bound_score, bound_score + 1, depth_next, var_stack);
			undo_move();
			if (temp_score == SCORE_INVAL) goto label_final;
			if (temp_score > bound_score)
			{
				jroot = iroot;
				root[jroot].root_score = bound_score = temp_score;
				my_printf("fh %d", depth);
#ifdef GUI
				print_cache(root[jroot].root_move, var_stack);
#else
				analyze_cache(root[jroot].root_move, var_stack);
#endif
				my_printf(" %.3f\n", (double) root[jroot].root_score);
#ifdef GUI
				print_best_move(depth, jroot);
#endif
				principle_variation = TRUE;
#ifdef GUI
				print_cur_move(depth, jroot);
#endif
				do_move(node->move_base[root[jroot].root_move]);
				fh_score = temp_score;
				temp_score = alpha_beta(temp_score, SCORE_LIMIT, depth_next, var_stack);
				undo_move();
				if (temp_score == SCORE_INVAL) goto label_final;
				principle_variation = FALSE;
				if (temp_score < fh_score) my_printf("research error!\n");
				root[jroot].root_score = bound_score = temp_score;
				my_printf("pv %d", depth);
#ifdef GUI
				print_cache(root[jroot].root_move, var_stack);
#else
				analyze_cache(root[jroot].root_move, var_stack);
#endif
				my_printf(" %.3f\n", (double) root[jroot].root_score);
#ifdef GUI
				print_best_move(depth, jroot);
#endif
			}
		}
		
		my_printf("depth=%d score=%.3f time=%.2f stats=%ld/%ld/%ld/%ld/%ld\n",
			depth, (double) root[jroot].root_score, time_used,
			ngens, ndos, nevals, nhits, nextends);
#ifndef GUI
		if (depth == DepthMax)
		{
			BUG((f = fopen("spread.txt", "a")) == NULL)
			fprintf(f, "%.2f\t%ld\t%ld\n", time_used, nodes_tot, nevals_tot);
			BUG(fclose(f) == EOF)
		}
#endif
		if (ABS(root[jroot].root_score) > SCORE_ABORT) goto label_final;
	}
	label_final:
	if (temp_score == SCORE_INVAL)
	{
		my_printf("time limit depth=%d move=%d time=%.2f stats=%ld/%ld/%ld/%ld/%ld\n",
			depth, iroot, time_used, ngens, ndos, nevals, nhits, nextends);
	}
	BUG(null_move != FALSE)
	BUG(nthreats != 0)
	if (jroot == INVALID) jroot = 0;
}

board_t *endgame_list[LIST_MAX];
int endgame_where[LIST_MAX];
int nendgame;
int endgame_example;

void set_endgame_lists(char *, char *);
void set_endgame_lists(char *wchar, char *bchar)
{
	int i;
	int icolour, ikind;
	board_t *p;
	char *s;
	
	for (i = 0; i < BOARD_MAX; i++)
	{
		board[i] = &side;
		asm_board[i] = 0;
	}
	for (i = 0; i < 50; i++)
		board[map[i]] = &empty;
	
	nendgame = 0;
	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		tot_count[icolour] = 0;
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			p = list[icolour][ikind];
			for (i = 0; i < LIST_MAX; i++)
			{
				p->board_colour = icolour;
				p->board_kind = INVALID;
				p->board_value = value[ikind];
				p->board_where = INVALID;
				p->board_key = hash_key[icolour][ikind];
				p->board_dir = dir[icolour][ikind];
				p->board_count = piece_count[icolour] + ikind;
				p->board_pat = pat[icolour][ikind];
				p++;
			}
			p = list[icolour][ikind];
			piece_count[icolour][ikind] = 0;
			if (icolour == WHITE_COLOUR)
				s = wchar;
			else
				s = bchar;
			while (*s != '\0')
			{
				if (*s == nota_kind[ikind])
				{
					piece_count[icolour][ikind]++;
					p->board_kind = ikind;
					p->board_where = INVALID;
					BUG(nendgame >= LIST_MAX);
					endgame_list[nendgame] = p;
					nendgame++;
					p++;
				}
				s++;
			}
			tot_count[icolour] += piece_count[icolour][ikind];
		}
	}
	
	node = node_stack;
	node->move_base = move_stack;
	node->stack_base = stack;
	node->node_key = return_cur_key();
	node->node_total_key = node->node_key;
	node->node_mat[WHITE_COLOUR] = return_mat_score(WHITE_COLOUR);
	node->node_mat[BLACK_COLOUR] = return_mat_score(BLACK_COLOUR);
	node->node_reverse = FALSE;
	null_move = FALSE;
	nthreats = 0;
	my_colour = WHITE_COLOUR;
	your_colour = BLACK_COLOUR;
}

int valid_endgame(void);
int valid_endgame(void)
{
	int iendgame, jendgame;
	
	for (iendgame = 0; iendgame < nendgame; iendgame++)
	{
		for (jendgame = iendgame + 1; jendgame < nendgame; jendgame++)
		{
			if (endgame_where[iendgame] == endgame_where[jendgame]) return(FALSE);
		}
	}
	
	for (iendgame = 0; iendgame < nendgame; iendgame++)
	{
		if (endgame_list[iendgame]->board_kind == MAN)
		{
			if ((endgame_list[iendgame]->board_colour == WHITE_COLOUR) and 
			    (endgame_where[iendgame] < 5)) return(FALSE);
			if ((endgame_list[iendgame]->board_colour == BLACK_COLOUR) and 
			    (endgame_where[iendgame] > 44)) return(FALSE);
		}
	}
	return(TRUE);
}

void set_endgame_board(void);
void set_endgame_board(void)
{
	int iendgame;
	board_t *p;
	int i;
#if DEBUG_LEVEL > 0
	int err;
#endif
	
	node->node_key.lo = node->node_key.hi = 0;
	node->node_mat[WHITE_COLOUR] = node->node_mat[BLACK_COLOUR] = 0;
	for (iendgame = 0; iendgame < nendgame; iendgame++)
	{
		p = endgame_list[iendgame];
		i = endgame_where[iendgame];
		p->board_where = map[i];
		board[map[i]] = p;
		asm_board[map[i]] = p->board_pat;
		UPDATE_KEY(node->node_key, p->board_key[p->board_where])
		node->node_mat[p->board_colour] += p->board_value;
	}
	node->node_total_key = node->node_key;

#if DEBUG_LEVEL > 0			
		if ((err = check_list()) > 0)
		{
			my_printf("err=%d\n", err);
			display_board();
			display_list();
			FATAL_ERROR("list error")
		}
		if ((err = check_asm_board()) > 0)
		{
			my_printf("err=%d\n", err);
			display_board();
			display_list();
			FATAL_ERROR("asm error")
		}
#endif
}

void clear_endgame_board(void);
void clear_endgame_board(void)
{
	int iendgame;
	board_t *p;
	int i;
	
	for (iendgame = 0; iendgame < nendgame; iendgame++)
	{
		p = endgame_list[iendgame];
		i = endgame_where[iendgame];
		p->board_where = INVALID;
		board[map[i]] = &empty;
		asm_board[map[i]] = 0;
	}
}

#define NSTATS 200

void analyze_endgame(char *, char *);
void analyze_endgame(char *wchar, char *bchar)
{
	long i;
	int j;
	long stats[NSTATS];
#if MACHINE != DOS
	endgame_t *dbase;
#endif
	int m;
	long n;
	FILE *f;
	char fname[LINE_MAX];
	
	set_endgame_lists(wchar, bchar);
	m = 0;
	m ^= wc_key[piece_count[WHITE_COLOUR][CROWN]];
	m ^= wm_key[piece_count[WHITE_COLOUR][MAN]];
	m ^= bc_key[piece_count[BLACK_COLOUR][CROWN]];
	m ^= bm_key[piece_count[BLACK_COLOUR][MAN]];
#if MACHINE == DOS
	XMShandle = endgame_dbase[m];
#else
	dbase = endgame_dbase[m];
#endif
	n = endgame_size[m];
	my_printf("stats %s %s %d %ld\n", wchar, bchar, m, n);
	
	my_printf("stats\n");
	for (j = 0; j < NSTATS; j++)
		stats[j] = 0;
	for (i = 0; i < n; i++)
	{
#if MACHINE == DOS
		XMSoffset = i;
		j = get_endgame();
#else
		j = dbase[i];
#endif
		if (j > 0) stats[j]++;
	}
	for (j = 0; j < NSTATS; j++)
		if (stats[j] > 0) my_printf("%d %ld\n", j, stats[j]);
	for (j = 0; j < NSTATS; j++)
		stats[j] = 0;
	for (i = 0; i < n; i++)
	{
#if MACHINE == DOS
		XMSoffset = i;
		j = get_endgame();
#else
		j = dbase[i];
#endif
		if (j <= 0) stats[-j]++;
	}
	for (j = 0; j < NSTATS; j++)
		if (stats[j] > 0) my_printf("%d %ld\n", -j, stats[j]);
	sprintf(fname, "%sv%s.bin", wchar, bchar);
	my_printf("writing %s\n", fname);
	BUG((f = fopen(fname, "wb")) == NULL)
#if MACHINE == DOS
	HUGE_FWRITE(endgame_dbase[m], endgame_t, n, f)
#else
	BUG(fwrite(endgame_dbase[m], sizeof(endgame_t), n, f) != n)
#endif
	FCLOSE(f)
}

int is_won(int);
int is_won(int mate)
{
	int imove;
	int nmoves;
	int escape;
	int temp;

	BUG(my_colour != BLACK_COLOUR)
	BUG(your_colour != WHITE_COLOUR)
	node->move_base = move_top;
	node->stack_base = stack_top;
	nmoves = gen_moves();
	escape = FALSE;
	for (imove = 0; imove < nmoves; imove++)
	{
		do_move(node->move_base[imove]);
		set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
		BUG(endgame_pointer == NULL)
		temp = get_endgame();
		if (temp == ENDGAME_DRAW)
			escape = TRUE;
		else if (temp <= 0)
			escape = TRUE;
		else if (temp > mate)
			escape = TRUE;
		undo_move();
		if (escape) break;
	}
	move_top = node->move_base;
	stack_top = node->stack_base;
	if (escape)
		return(FALSE);
	else
		return(TRUE);
}

void make_complete(int);
void make_complete(int mate)
{
	int imove;
	int nmoves;
	int temp;
	
	BUG(my_colour != WHITE_COLOUR)
	BUG(your_colour != BLACK_COLOUR)
	node->move_base = move_top;
	node->stack_base = stack_top;
	nmoves = gen_inverse_moves();
	for (imove = 0; imove < nmoves; imove++)
	{
		do_move(node->move_base[imove]);
		if (can_capture(WHITE_COLOUR, BLACK_COLOUR))
		{
			undo_move();
			continue;
		}
		set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
		BUG(endgame_pointer == NULL)
		temp = get_endgame();
		if (temp == ENDGAME_DRAW)
		{
			put_endgame(mate + 2);
		}
		undo_move();
	}
	move_top = node->move_base;
	stack_top = node->stack_base;
}

void gen_won_endgame(char *, char *);
void gen_won_endgame(char *wchar, char *bchar)
{
#define INIT\
	for (iendgame = 0; iendgame < nendgame; iendgame++)\
		endgame_where[iendgame] = 0;\
	kendgame = nendgame - 1;

#define CONTINUE\
		while(kendgame >= 0)\
		{\
			endgame_where[kendgame]++;\
			if (endgame_where[kendgame] < 50) break;\
			kendgame--;\
		}\
		if (kendgame < 0) break;\
		while(kendgame < (nendgame - 1))\
		{\
			kendgame++;\
			endgame_where[kendgame] = 0;\
		}

	int iendgame, kendgame;
	int nwhite_moves, nblack_moves;
	int imove;
	int mate, best_mate;
	int ready;

	set_endgame_lists(wchar, bchar);
	INIT
	while(TRUE)
	{
		if (valid_endgame())
		{
			set_endgame_board();
			set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
			BUG(endgame_pointer == NULL)
			mate = get_endgame();
			
			if (mate == ENDGAME_DRAW)
			{
				node->move_base = move_top;
				node->stack_base = stack_top;
				nwhite_moves = gen_moves();
				if (nwhite_moves == 0)
				{
					display_board();
					my_printf("lost for white in 0!\n");
					put_endgame(0);
				}
				else
				{
					best_mate = ENDGAME_INVAL;
					for (imove = 0; imove < nwhite_moves; imove++)
					{
						do_move(node->move_base[imove]);
						set_endgame_pointer(BLACK_COLOUR, WHITE_COLOUR);
						mate = get_endgame();
						if (mate == ENDGAME_DRAW)
						{
							node->move_base = move_top;
							node->stack_base = stack_top;
							nblack_moves = gen_moves();
							move_top = node->move_base;
							stack_top = node->stack_base;
							if (nblack_moves == 0)
								mate = 1;
						}
						else
						{
							if (mate > 0)
								mate = -mate - 1;
							else
								mate = -mate + 1;
						}
						if (best_mate == ENDGAME_INVAL)
							best_mate = mate;
						else
						{
							if (best_mate > 0)
							{
								if ((mate > 0) and (mate < best_mate))
									best_mate = mate;
							}
							else
							{
								if (mate > 0)
									best_mate = mate;
								else if (mate < best_mate)
									best_mate = mate;
							}
						}
						undo_move();
						if (best_mate == 1) break;
					}
					set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
					BUG(endgame_pointer == NULL)
					put_endgame(best_mate);
				}
				move_top = node->move_base;
				stack_top = node->stack_base;
			}
			clear_endgame_board();
		}
		CONTINUE
	}
	
	mate = 1;
	while(TRUE)
	{
		my_printf("trying %d\n", mate);
		ready = TRUE;
		endgame_example = TRUE;
		INIT
		while(TRUE)
		{
			if (valid_endgame())
			{
				set_endgame_board();
				set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
				BUG(endgame_pointer == NULL)								
				if (get_endgame() == mate)
				{
					if (endgame_example)
					{
						display_board();
						my_printf("this is won in %d\n", mate);
						endgame_example = FALSE;
					}
					ready = FALSE;
					node->move_base = move_top;
					node->stack_base = stack_top;
					my_colour = BLACK_COLOUR;
					your_colour = WHITE_COLOUR;
					nblack_moves = gen_inverse_moves();
					for (imove = 0; imove < nblack_moves; imove++)
					{
						do_move(node->move_base[imove]);
						if (can_capture(BLACK_COLOUR, WHITE_COLOUR))
						{
							undo_move();
							continue;
						}
						my_colour = BLACK_COLOUR;
						your_colour = WHITE_COLOUR;
						if (is_won(mate))
						{
							my_colour = WHITE_COLOUR;
							your_colour = BLACK_COLOUR;
							make_complete(mate);
						}
						my_colour = WHITE_COLOUR;
						your_colour = BLACK_COLOUR;
						undo_move();
					}
					move_top = node->move_base;
					stack_top = node->stack_base;
				}
				clear_endgame_board();
			}
			CONTINUE
		}
		if (ready) break;
		mate += 2;
	}
}

int is_lost(int);
int is_lost(int mate)
{
	int imove;
	int nmoves;
	int escape;
	int temp;
	
	BUG(my_colour != WHITE_COLOUR)
	BUG(your_colour != BLACK_COLOUR)
	node->move_base = move_top;
	node->stack_base = stack_top;
	nmoves = gen_moves();
	escape = FALSE;
	for (imove = 0; imove < nmoves; imove++)
	{
		do_move(node->move_base[imove]);
		set_endgame_pointer(BLACK_COLOUR, WHITE_COLOUR);
		BUG(endgame_pointer == NULL)
		temp = get_endgame();
		if (temp == ENDGAME_DRAW)
			escape = TRUE;
		else if (temp <= 0)
			escape = TRUE;
		else if (temp > mate)
			escape = TRUE;
		undo_move();
		if (escape) break;
	}
	move_top = node->move_base;
	stack_top = node->stack_base;
	if (escape)
		return(FALSE);
	else
		return(TRUE);
}

void gen_lost_endgame(char *, char *);
void gen_lost_endgame(char *wchar, char *bchar)
{
#define INIT\
	for (iendgame = 0; iendgame < nendgame; iendgame++)\
		endgame_where[iendgame] = 0;\
	kendgame = nendgame - 1;

#define CONTINUE\
		while(kendgame >= 0)\
		{\
			endgame_where[kendgame]++;\
			if (endgame_where[kendgame] < 50) break;\
			kendgame--;\
		}\
		if (kendgame < 0) break;\
		while(kendgame < (nendgame - 1))\
		{\
			kendgame++;\
			endgame_where[kendgame] = 0;\
		}

	int iendgame, kendgame;
	int nwhite_moves;
	int imove;
	int ready;
	int mate;
	int temp;

	set_endgame_lists(wchar, bchar);

	mate = 1;
	while(TRUE)
	{
		ready = TRUE;
		endgame_example = TRUE;
		my_printf("trying %d\n", mate);
		INIT
		while(TRUE)
		{
			if (valid_endgame())
			{
				set_endgame_board();
				if (endgame_example)
				{
					set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
					BUG(endgame_pointer == NULL)
					temp = get_endgame();			
					if (temp == (-mate - 1))
					{
						display_board();
						my_printf("this is lost in %d\n", mate + 1);
						endgame_example = FALSE;
					}
				}
				set_endgame_pointer(BLACK_COLOUR, WHITE_COLOUR);
				BUG(endgame_pointer == NULL)
				temp = get_endgame();			
				if (temp == mate)
				{
					ready = FALSE;
					
					node->move_base = move_top;
					node->stack_base = stack_top;
					nwhite_moves = gen_inverse_moves();
					for (imove = 0; imove < nwhite_moves; imove++)
					{
						do_move(node->move_base[imove]);
						if (can_capture(WHITE_COLOUR, BLACK_COLOUR))
						{
							undo_move();
							continue;
						}
						my_colour = WHITE_COLOUR;
						your_colour = BLACK_COLOUR;
						set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
						BUG(endgame_pointer == NULL)
						temp = get_endgame();
						if (temp == ENDGAME_DRAW)
						{
							if (is_lost(mate))
							{
								set_endgame_pointer(WHITE_COLOUR, BLACK_COLOUR);
								BUG(endgame_pointer == NULL)
								put_endgame(-mate - 1);
								if (endgame_example)
								{
									display_board();
									my_printf("this is lost in %d\n", -mate - 1);
									endgame_example = FALSE;
								}
							}
						}
						my_colour = BLACK_COLOUR;
						your_colour = WHITE_COLOUR;
						undo_move();
					}
					move_top = node->move_base;
					stack_top = node->stack_base;
				}
				clear_endgame_board();
			}
			CONTINUE
		}
		if (ready) break;
		mate += 2;
	}
}

void do_endgame(void)
{
	FILE *f;
	char wchar[LINE_MAX], bchar[LINE_MAX];
	int option;
	
	BUG((f = fopen("endgame.dat", "r")) == NULL)
	while(fscanf(f, "%s%s%d", wchar, bchar, &option) == 3)
	{
		if (option == 0) continue;
		my_printf("generating won positions for %s %s\n", wchar, bchar);
		gen_won_endgame(wchar, bchar);
	}
	FCLOSE(f)
	BUG((f = fopen("endgame.dat", "r")) == NULL)
	while(fscanf(f, "%s%s%d", wchar, bchar, &option) == 3)
	{
		if (option == 0) continue;
		my_printf("generating lost positions for %s %s\n", wchar, bchar);
		gen_lost_endgame(wchar, bchar);
		analyze_endgame(wchar, bchar);
	}
	FCLOSE(f)
}

void minimax(void)
{
	int temp_score, best_score;
	int imove, nmoves;
	int best_move;
	int temp;
	
	display_board();
	set_endgame_pointer(my_colour, your_colour);
	temp = get_endgame();
	if (temp == ENDGAME_DRAW)
	{
		my_printf("%sdraws\n", colour_as_string[my_colour]);
	}
	else if (temp > 0)
	{
		my_printf("%swins in %d moves\n", colour_as_string[my_colour], temp);
	}
	else
	{
		my_printf("%sloses in %d moves\n", colour_as_string[my_colour], -temp);
	}
	
	/*
	select best move
	*/
	
	best_score = -SCORE_LIMIT;
	best_move = INVALID;
	node->move_base = move_top;
	node->stack_base = stack_top;
	nmoves = gen_moves();
	for (imove = 0; imove < nmoves; imove++)
	{
		display_move("", node->move_base[imove], "");
		do_move(node->move_base[imove]);
		set_endgame_pointer(my_colour, your_colour);
		temp = get_endgame();
		if (temp == ENDGAME_DRAW)
			temp_score = 0;
		else if (temp > 0)
			temp_score = SCORE_ENDGAME - temp;
		else
			temp_score = -SCORE_ENDGAME - temp;
		temp_score = -temp_score;
		my_printf(" score=%d\n", temp_score);
		if (temp_score > best_score)
		{
			best_move = imove;
			best_score = temp_score;
		}
		undo_move();
	}
	if (best_move != INVALID)
	{
		display_move("best move=", node->move_base[best_move], "");
		my_printf(" best score=%d\n", best_score);
		do_move(node->move_base[best_move]);
		minimax();
		undo_move();
	}
	move_top = node->move_base;
	stack_top = node->stack_base;
}

#if MACHINE == DOS
#define XMS_ALLOC(H, N)\
{\
	my_printf("XMS alloc %s Kbytes=%lu\n", #H, (ulong_t) (N));\
	BUG(((H) = XMS_alloc(N)) == 0);\
}
#define XMSFREE(H) {XMS_free(H);}

void XMSCLEAR(unsigned int handle, int kbytes)
{
	int i;
	char buffer[KB];
	struct XMSmovestruct ms;

	memset(buffer, ENDGAME_DRAW, KB);
	ms.nbytes = KB;
	ms.srchandle = 0;
	ms.srcoffset = (long) buffer;
	ms.dsthandle = handle;
	ms.dstoffset = 0;
	for (i = 0; i < kbytes; i++)
	{
		XMS_move(&ms);
		ms.dstoffset += KB;
	}
}
#endif

void init(void)
{
	int i;
	int icolour, ikind;
	FILE *f, *g;
	char wchar[LINE_MAX], bchar[LINE_MAX], fname[LINE_MAX];
	int m;
	long n;
	long j;
#if MACHINE == DOS
	int kbytes;
#endif
	int option;
	
	for (i = 0; i < 50; i++)
		BUG(inverse_map[map[i]] != i)
	
	MALLOC(row, int, BOARD_MAX)
	for (i = 0; i < BOARD_MAX; i++)
		row[i] = -1;
	for (i = 0; i < 50; i++)
		row[map[i]] = row_table[i];
	MALLOC(move_stack, move_t *, MOVE_MAX)
	MALLOC(stack, move_t, STACK_MAX)
	MALLOC(capt_stack, board_t *, CAPT_MAX)
	MALLOC(var_stack, var_t, VAR_MAX)
	move_top = move_stack;
	stack_top = stack;
	capt_top = capt_stack;
	MALLOC(node_stack, node_t, NODE_MAX)
	node_stack[0].node_reverse = FALSE;
	node_stack++;
	node = node_stack;
	inode = 0;
	
	my_printf("size of hash tables: %lu %lu\n",
		(ulong_t) sizeof(cache_t) * CACHE_MASK * 2,
		(ulong_t) sizeof(score_cache_t) * SCORE_CACHE_MASK * 2);
#if MACHINE == DOS
	HUGE_MALLOC(cache[WHITE_COLOUR], cache_t, CACHE_MASK + 1)
	HUGE_MALLOC(cache[BLACK_COLOUR], cache_t, CACHE_MASK + 1)
	HUGE_MALLOC(score_cache[WHITE_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
	HUGE_MALLOC(score_cache[BLACK_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
#else
	MALLOC(cache[WHITE_COLOUR], cache_t, CACHE_MASK + 1)
	MALLOC(cache[BLACK_COLOUR], cache_t, CACHE_MASK + 1)
	MALLOC(score_cache[WHITE_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
	MALLOC(score_cache[BLACK_COLOUR], score_cache_t, SCORE_CACHE_MASK + 1)
#endif

	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			MALLOC(hash_key[icolour][ikind], hash_key_t, BOARD_MAX)
			for (i = 0; i < 50; i++)
			{
				hash_key[icolour][ikind][map[i]].lo = ulrand();
				hash_key[icolour][ikind][map[i]].hi = ulrand();
			}
		}
	}
	MALLOC(threat_key, hash_key_t, NODE_MAX)
	for (i = 0; i < NODE_MAX; i++)
	{
		threat_key[i].lo = ulrand();
		threat_key[i].hi = ulrand();
	}
	
	MALLOC(white_man_visit, int, BOARD_MAX)
	MALLOC(black_man_visit, int, BOARD_MAX)
	man_visit[WHITE_COLOUR] = white_man_visit;
	man_visit[BLACK_COLOUR] = black_man_visit;
	for (i = 0; i < BOARD_MAX; i++)
	{
		white_man_visit[i] = black_man_visit[i] = 0;
	}
	for (i = 0; i < 50; i++)
	{
		white_man_visit[map[i]] = man_visit_table[i];
		black_man_visit[mirror_map[i]] = man_visit_table[i];
	}
	
	for (i = 0; i < 51; i++)
		bempty[i] = 0;
	
	MALLOC(root, root_t, NROOT_MAX)
	MALLOC(game_moves, char, NGAME_MAX)
	
	wc_key[0] = wm_key[0] = bc_key[0] = bm_key[0] = 0;
	for (i = 1; i < ENDGAME_MAX; i++)
	{
		wc_key[i] = (int) (ulrand() & ENDGAME_MASK);
		wm_key[i] = (int) (ulrand() & ENDGAME_MASK);
		bc_key[i] = (int) (ulrand() & ENDGAME_MASK);
		bm_key[i] = (int) (ulrand() & ENDGAME_MASK);
	}
	for (i = 0; i < ENDGAME_SIZE; i++)
		endgame_dbase[i] = NULL;
	f = fopen("endgame.dat", "r");
	BUG(f == NULL)
	while(fscanf(f, "%s%s%d", wchar, bchar, &option) == 3)
	{
		my_printf("read %s %s\n", wchar, bchar);
		set_endgame_lists(wchar, bchar);
		m = 0;
		m ^= wc_key[piece_count[WHITE_COLOUR][CROWN]];
		m ^= wm_key[piece_count[WHITE_COLOUR][MAN]];
		m ^= bc_key[piece_count[BLACK_COLOUR][CROWN]];
		m ^= bm_key[piece_count[BLACK_COLOUR][MAN]];
		BUG(m > ENDGAME_MASK)
		BUG(m < 0)
		n = 1;
		for (i = 0; i < WTOT + BTOT; i++)
			n *= 50L;
		my_printf("%s %s %d %ld\n", wchar, bchar, m, n);
		BUG(endgame_dbase[m] != NULL)
		endgame_size[m] = n;
#if MACHINE == DOS
		BUG(!XMS_installed())
		kbytes = (int) ((((n) * sizeof(endgame_t)) + KB - 1) / KB);
		XMS_ALLOC(endgame_dbase[m], kbytes);
		XMSCLEAR(endgame_dbase[m], kbytes);
		c2x.nbytes = 4;
		c2x.srchandle = 0;
		c2x.srcoffset = (long) XMSresult;
		c2x.dsthandle = 0;
		c2x.dstoffset = 0;
	
		x2c.nbytes = 4;
		x2c.srchandle = 0;
		x2c.srcoffset = 0;
		x2c.dsthandle = 0;
		x2c.dstoffset = (long) XMSresult;
#else
		MALLOC(endgame_dbase[m], endgame_t, n);
#endif
		sprintf(fname, "%sv%s.bin", wchar, bchar);
		if ((g = fopen(fname, "rb")) == NULL)
		{
			my_printf("could not read %s!\n", fname);
#if MACHINE != DOS
			for (j = 0; j < n; j++)
				endgame_dbase[m][j] = ENDGAME_DRAW;
#endif
		}
		else
		{
#if MACHINE == DOS
			XMS_FREAD(endgame_dbase[m], endgame_t, n, g)
#else
			BUG(fread(endgame_dbase[m], sizeof(endgame_t), n, g) != n)
#endif
			FCLOSE(g)
		}
	}
	FCLOSE(f)
}

void quit(void)
{
	int icolour, ikind;
	int i;
	
	FREE(row)
	FREE(move_stack)
	FREE(stack)
	FREE(capt_stack)
	FREE(var_stack)
	FREE(node_stack - 1)
#if MACHINE == DOS
	HUGE_FREE(cache[WHITE_COLOUR])
	HUGE_FREE(cache[BLACK_COLOUR])
	HUGE_FREE(score_cache[WHITE_COLOUR])
	HUGE_FREE(score_cache[BLACK_COLOUR])
#else
	FREE(cache[WHITE_COLOUR])
	FREE(cache[BLACK_COLOUR])
	FREE(score_cache[WHITE_COLOUR])
	FREE(score_cache[BLACK_COLOUR])
#endif

	for (icolour = WHITE_COLOUR; icolour <= BLACK_COLOUR; icolour++)
	{
		for (ikind = MAN; ikind <= CROWN; ikind++)
		{
			FREE(hash_key[icolour][ikind])
		}
	}
	FREE(threat_key)
	FREE(white_man_visit)
	FREE(black_man_visit)
	FREE(root)
	FREE(game_moves)
	
	for (i = 0; i < ENDGAME_SIZE; i++)
	{
		if (endgame_dbase[i] != NULL)
#if MACHINE == DOS
			XMS_FREE(endgame_dbase[i])
#else
			FREE(endgame_dbase[i])
#endif
	}
	
	my_printf("%z");
}

#ifdef GUI
#define START_BOARD "brd.txt"
move_t *get_move(int WhiteThink)
{
	FILE *f;

	BUG((f = fopen(START_BOARD, "r")) == NULL)
	BUG(!read_board(f))
	BUG(fclose(f) == EOF)

	if (WhiteThink)
	{
		my_colour = WHITE_COLOUR;
		your_colour = BLACK_COLOUR;
	}
	else
	{
		my_colour = BLACK_COLOUR;
		your_colour = WHITE_COLOUR;
	}
	gen_root();
	if (nroot == 0)
		return (NULL);
	if (nroot > 1)
		do_root();
	if (node != node_stack)
		my_printf("node != node_stack!\n");
	return (node->move_base[root[jroot].root_move]);
}

int check_time(void)
{
	gwdupdatetime();
	if (gwdabortthink())
		return (1);
	if (time_used > TimeLimit)
	{
		if ((TimeExtra > 0) && ((begin_score - root[jroot].root_score) > MARGIN))
		{
			TimeLimit += TimeExtra;
			TimeExtra = 0;
		}
	}
	if (time_used > TimeLimit)
		return (1);
	return (0);
}

void print_move(char *move, char *string)
{
	sprintf(string, "%02d%c%02d",
		pat_map[*move], strlen(move) > 2 ? '*' : '-', pat_map[*(move+1)]);
}

void print_cur_move(int depth, int iroot)
{
	char string[13];
	char move[6];
	print_move(node->move_base[root[iroot].root_move], move);
	sprintf(string, "%2d.%02d  %s", depth, iroot + 1, move);
	gwddrawcurrentmove(string);
}

void print_best_move(int depth, int iroot)
{
	char string[13];
	sprintf(string, "%2d.%02d %6ld", depth, iroot + 1, (long) root[iroot].root_score);
	gwddrawbestmove(string);
}

void print_cache(move_t best_move, var_t *var_base)
{
	gwdclearpv();
	analyze_cache(best_move, var_base);
	gwddrawpv();
}

void print_cache_move(char *move)
{
	char string[6];
	print_move(move, string);
	gwdaddpv(string);
}
#endif
